/*
 * onyx-the-black-cat.c
 * Fractal Guru <reverse@put.as> http://reverse.put.as
 *
 * For use with Mac OS X Tiger v10.4.11 and Leopard 10.5.5
 *
 * Based on original code by Landon J. Fuller <landonf@opendarwin.org>
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define VERSION 0.2

/* COMPILE OPTIONS */
// change to 0 to remove monitoring of extended attributes calls
#define MONITOR_EXTENDED_ATTRIBUTES 1
// change to 0 to remove hex dump of attributes
#define DUMP_EXTENDED_ATTRIBUTES 1
// change to 0 to build for Leopard (sysent table and proc struct are different for Leopard)
#define BUILDING_ON_TIGER 1

// we need this to use the kinfo_proc structure for sysctl
#define __APPLE_API_UNSTABLE
#define SYSCTL_DEF_ENABLED
#define PROC_DEF_ENABLED

#include <mach/mach_types.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <stdbool.h> 
#include <sys/param.h>
#include <sys/types.h> 
#include <sys/proc.h>
#include <sys/sysctl.h>
#include <sys/xattr.h>
#include <stdint.h>
#include "structures.h"

static struct sysent *_sysent;
extern int nsysent;

typedef int	ptrace_func_t (struct proc *, struct ptrace_args *, int *);
typedef int	open_func_t   (struct proc *, struct open_args *, int *);
typedef int sysctl_func_t (struct proc *, struct sysctl_args *, int *);
typedef int chmod_func_t  (struct proc *, struct chmod_args *, int *);
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
typedef int getxattr_func_t (struct proc *, struct getxattr_args *, int *);
typedef int fgetxattr_func_t (struct proc *, struct fgetxattr_args *, int *);
typedef int setxattr_func_t (struct proc *, struct setxattr_args *, int *);
typedef int fsetxattr_func_t (struct proc *, struct fsetxattr_args *, int *);
typedef int removexattr_func_t (struct proc *, struct removexattr_args *, int *);
typedef int fremovexattr_func_t (struct proc *, struct fremovexattr_args *, int *);
typedef int listxattr_func_t (struct proc *, struct listxattr_args *, int *);
typedef int flistxattr_func_t (struct proc *, struct flistxattr_args *, int *);
#endif

static ptrace_func_t *real_ptrace;
static open_func_t *real_open;
static sysctl_func_t *real_sysctl;
static chmod_func_t *real_chmod;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
static getxattr_func_t *real_getxattr;
static fgetxattr_func_t *real_fgetxattr;
static setxattr_func_t *real_setxattr;
static fsetxattr_func_t *real_fsetxattr;
static removexattr_func_t *real_removexattr;
static fremovexattr_func_t *real_fremovexattr;
static listxattr_func_t *real_listxattr;
static flistxattr_func_t *real_flistxattr;
#endif

// Hijack ptrace syscall
static int our_ptrace (struct proc *p, struct ptrace_args *uap, int *retval)
{
	if (uap->req == PT_DENY_ATTACH) {
		printf("[onyx-the-black-cat] Blocking PT_DENY_ATTACH for pid %d.\n", uap->pid);
		return (0);
	} else {
		return real_ptrace(p, uap, retval);
	}
}

// Hijack the open syscall
static int our_open (struct proc *p, struct open_args *uap, int *retval)
{
 int error;
 size_t dummy = 0;
 char savedpath[MAXPATHLEN];
// copy from user space to kernel space
//    copyinstr()   Copies a NUL-terminated string, at most len bytes long, from user-space address uaddr to
//    kernel-space address kaddr.  The number of bytes actually copied, including the terminating 
//	  NUL, is returned in *done.
//	  int copyinstr(const void *uaddr, void *kaddr, size_t len, size_t *done);
	error = copyinstr(uap->path, (void *)savedpath, MAXPATHLEN, &dummy);
    if (!error)
        printf("open(%s, %x, %x)\n", savedpath, uap->flags, uap->mode);
    else
        printf("open(?, %x, %x)\n", uap->flags, uap->mode);

    return real_open(p, uap, retval);
}

/*
The system call function implements the system call. Its function prototype is 
defined in the <sys/sysent.h> header as: 
typedef int32_t sy_call_t(struct proc *, void *, int *);
where struct proc * points to the currently calling process, and void * 
points to the system call’s arguments’ structure, if there is any, and
int * to the return value. 
*/
// bsd/vfs/vfs_syscalls.c
/* find the function that implements the syscal at bsd/kern/init_sysent.c */

// Hijack chmod syscall
static int our_chmod (struct proc *p, struct chmod_args *uap, int *retval)
{
 int error;
 size_t dummy = 0;
 char savedpath[MAXPATHLEN];
 //      chmod(const char *path, mode_t mode);
 error = copyinstr(uap->path, savedpath, MAXPATHLEN, &dummy);
 if (!error)
	printf("chmod requested to filename: %s\n",savedpath);
 else
	printf("chmod requested to unknown filename\n");
 // pass control to the real chmod function
 return real_chmod(p, uap, retval);
}

// our sysctl so we can intercept the anti-debug call
/*
 function that implements the sysctl syscall
         {AC(__sysctl_args), _SYSCALL_CANCEL_NONE, KERNEL_FUNNEL, (sy_call_t *)__sysctl, munge_wwwwww, munge_dddddd, _SYSCALL_RET_INT_T}, // 202 = __sysctl
 so it's "__sysctl"
*/
// __sysctl(struct proc *p, struct __sysctl_args *uap, __unused register_t *retval)
// retval is unused
// sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp, size_t newlen);
// proc structure is defined in sys/proc_internal.h
// Hijack __sysctl
static int our_sysctl (struct proc *p, struct sysctl_args *uap, int *retval)
{
	int mib[4];
	int result;
	struct kinfo_proc kpr;
	
	// call the real_sysctl function and hold the result so we can use it to return from our hijacked sysctl
	result = real_sysctl(p,uap,retval);
 
 /* 
 **************************************************
  How to bypass the anti-debug protection:
  If we copy the kinfo_proc structure to kernel space and edit the p_flag and copy back to userspace,
  then we can bypass the anti-debug protection 
 ***************************************************
 */
	// int copyin(const void *uaddr, void *kaddr, size_t len); 
	// copyin() Copies len bytes of data from the user-space address uaddr to the kernel-space address kaddr.
	// copy structure from userspace to kernel space so we can verify if it's what we are looking for
	copyin(uap->name, &mib, sizeof(mib));
	// if it's a anti-debug call
	if (mib[0]==CTL_KERN && mib[1]==KERN_PROC && mib[2]==KERN_PROC_PID) 
	{
		// then copy the result from the destination buffer ( *oldp from sysctl call) to kernel space so we can edit
		copyin(uap->old,&kpr,sizeof(kpr));
		// we can display the PID from the calling program, which can be useful 
		printf("[onyx-the-black-cat] Detected possible sysctl anti-debug trick requested by process with PID %d (%s)! Patching...\n", kpr.kp_proc.p_pid, kpr.kp_proc.p_comm);
		// modify the p_flag because:
		// We're being debugged if the P_TRACED flag is set.
		// return ( (info.kp_proc.p_flag & P_TRACED) != 0 );
		kpr.kp_proc.p_flag = 0x0;
		// copy back to user space the modified structure
		// int copyout(const void *kaddr, void *uaddr, size_t len);
		// copyout() Copies len bytes of data from the kernel-space address kaddr to the user-space address uaddr.
		copyout(&kpr, uap->old,sizeof(kpr));
   /* sysctl()
     If the old value is not desired, oldp and
     oldlenp should be set to NULL.
   */
 }
 // and return from our hijacked function
 return result;
}

#if MONITOR_EXTENDED_ATTRIBUTES >= 1
// we can get the process name from p_comm field of a proc structure
// but it's size is limited so we don't get the full name
// we could use the proc_selfname or proc_name name functions but they aren't any better since they retrieve the same p_comm field
// Hijack getxattr
static int our_getxattr (struct proc *p, struct getxattr_args *uap, int *retval)
{
 int result, i=0, error;
 size_t attributenamelen, attributesize;
 char attributename[XATTR_MAXNAMELEN+1];

 attributesize = uap->size;
 char attributevalue[attributesize];
 // handle request for attribute size (value equal to NULL)
 if (uap->value == 0)
 { 
  result = real_getxattr(p, uap, retval);
 }
 else
 {
	 if (error = copyinstr(uap->attrname, attributename, sizeof(attributename), &attributenamelen) != 0)
	 {
		 printf("[onyx-the-black-cat] Error: Can't copy attribute name to kernel space!\n");
	 }
	 printf("[onyx-the-black-cat] Detected request to getxattr syscall from PID %d (%s) Name: %s Value: ",p->p_pid, p->p_comm, attributename);
	 result = real_getxattr(p, uap, retval);
	 copyin(uap->value, &attributevalue, sizeof(attributevalue));
	 printf("%s\n",attributevalue);
#if DUMP_EXTENDED_ATTRIBUTES >= 1
	 printf("[onyx-the-black-cat] getxattr Hex dump: ");
	 for (i=0; i<sizeof(attributevalue); i++) { printf("%x",attributevalue[i]); }
	 printf("\n");
#endif
 }
 return result;
}

// Hijack fgetxattr
// kernel implementation at xnu/bsd/vfs/vfs_syscalls.c
static int our_fgetxattr (struct proc *p, struct fgetxattr_args *uap, int *retval)
{
 int result, i=0, error;
 size_t attributenamelen, attributesize;
 char attributename[XATTR_MAXNAMELEN+1];

 attributesize = uap->size;
 char attributevalue[attributesize];
 /* man getxattr:
     On success, value contains the data associated with name.  When value is
     set to NULL, getxattr() returns current size of the named attribute.
     This facility can be used to determine the size of a buffer sufficiently
     large to hold the data currently associated with the attribute.
 */
 // we are just interested in dumping real reads to the attribute name
 if (uap->value == 0)
 {
	result = real_fgetxattr(p, uap, retval);
 }
 else
 {
	 // copy attribute name to kernel space
	 if(error = copyinstr(uap->attrname, attributename, sizeof(attributename), &attributenamelen) != 0)
	 {
		 printf("[onyx-the-black-cat] Error: Can't copy attribute name to kernel space!\n");
	 }
	 printf("[onyx-the-black-cat] Detected request to fgetxattr syscall from PID %d (%s) Attribute name: %s Value: ", p->p_pid, p->p_comm, attributename);
	 // call the real fgetxattr so we can have value
	 result = real_fgetxattr(p, uap, retval);
	 // copy value to kernel space
	 copyin(uap->value, &attributevalue, sizeof(attributevalue));
	 printf("%s\n", attributevalue);
#if DUMP_EXTENDED_ATTRIBUTES >= 1
	 printf("[onyx-the-black-cat] fgetxattr Hex dump: ");
	 for (i=0; i<sizeof(attributevalue); i++) { printf("%x",attributevalue[i]); }
	 printf("\n");
#endif
 }
 return result;
}

// Hijack setxattr
static int our_setxattr (struct proc *p, struct setxattr_args *uap, int *retval)
{
 int result;
 char attributepath[255];
 char attributename[XATTR_MAXNAMELEN+1];
 copyin(uap->path, &attributepath, sizeof(attributepath));
 copyin(uap->attrname, &attributename, sizeof(attributename));
 printf("[onyx-the-black-cat] Detected request to setxattr syscall from PID %d (%s) Path: %s Attribute name: %s\n", p->p_pid, p->p_comm, attributepath, attributename);
 result = real_setxattr(p, uap, retval);
 return result;
}

// Hijack fsetxattr
static int our_fsetxattr (struct proc *p, struct fsetxattr_args *uap, int *retval)
{
 int result;
 printf("[onyx-the-black-cat] Detected request to fsetxattr syscall from PID %d (%s)\n", p->p_pid, p->p_comm);
 result = real_fsetxattr(p, uap, retval);
 return result;
}

// Hijack removexattr
static int our_removexattr (struct proc *p, struct removexattr_args *uap, int *retval)
{
 int result;
 char attributepath[255];
 char attributename[XATTR_MAXNAMELEN+1];
 copyin(uap->path, &attributepath, sizeof(attributepath));
 copyin(uap->attrname, &attributename, sizeof(attributename));
 printf("[onyx-the-black-cat] Detected request to removexattr syscall from PID %d (%s) Path: %s Attribute name: %s\n", p->p_pid, p->p_comm, attributepath, attributename);
 result = real_removexattr(p, uap, retval);
 return result;
}

// Hijack fremovexattr
static int our_fremovexattr (struct proc *p, struct fremovexattr_args *uap, int *retval)
{
 int result;
 char attributename[XATTR_MAXNAMELEN+1];
 copyin(uap->attrname, &attributename, sizeof(attributename));
 printf("[onyx-the-black-cat] Detected request to fremovexattr syscall from PID %d (%s) Attribute name: %s\n", p->p_pid, p->p_comm, attributename);
 result = real_fremovexattr(p, uap, retval);
 return result;
}

// Hijack listxattr
static int our_listxattr (struct proc *p, struct listxattr_args *uap, int *retval)
{
 int result;
 char attributepath[255];
 if (uap->namebuf == 0)
 {
	result = real_listxattr(p, uap, retval);
 }
 else
 {
	 copyin(uap->path, &attributepath, sizeof(attributepath));
	 printf("[onyx-the-black-cat] Detected request to listxattr syscall from PID %d (%s) Path: %s\n", p->p_pid, p->p_comm, attributepath);
	 result = real_listxattr(p, uap, retval);
 }
 return result;
}

// Hijack flistxattr
static int our_flistxattr (struct proc *p, struct flistxattr_args *uap, int *retval)
{
 int result;
 if (uap->namebuf == 0)
 {
	result = real_flistxattr(p, uap, retval);
 }
 else
 {
	 printf("[onyx-the-black-cat] Detected request to flistxattr syscall from PID %d (%s)\n", p->p_pid, p->p_comm);
	 result = real_flistxattr(p, uap, retval);
 }
 return result;
}
#endif

/*
 * nsysent is placed directly before the hidden sysent, so skip ahead
 * and sanity check that we've found the sysent array.
 *
 * Clearly, this is extremely fragile and not for general consumption.
 */
static struct sysent *find_sysent () 
{
	unsigned int table_size;
	struct sysent *table;

	table_size = sizeof(struct sysent) * nsysent;
	table = (struct sysent *) ( ((char *) &nsysent) + sizeof(nsysent) );

#if __i386__
	/* For reasons unknown, the table is offset by an additional 28 bytes on my i386 system */
	table = (struct sysent *) ( ((uint8_t *) table) + 28);
#endif

	printf("[onyx-the-black-cat] Found nsysent at %p (count %d), calculated sysent location at %p.\n", &nsysent, nsysent, table);

	/* Sanity check */
	printf("[onyx-the-black-cat] Sanity check: verifying if number of syscalls arguments are the expected ones\n");
		
// number of arguments for each function (we can use xnu/bsd/sys/sysproto.h)
// if sysent table is found and it's correct, the number of arguments should match
	if (
		table[SYS_exit].sy_narg == 1  &&
		table[SYS_fork].sy_narg == 0 &&
		table[SYS_read].sy_narg == 3 &&
		table[SYS_wait4].sy_narg == 4 &&
		table[SYS_ptrace].sy_narg == 4 &&
		table[SYS_getxattr].sy_narg == 6 &&
		table[SYS_listxattr].sy_narg == 4)
	{
		printf("[onyx-the-black-cat] Sanity check: sysent sanity check succeeded.\n");
		return table;
	} else {
		printf("[onyx-the-black-cat] Sanity check: sanity check failed, could not find sysent table.\n");
		return NULL;
	}
}

// THE FUN STARTS HERE
kern_return_t onyx_the_black_cat_start (kmod_info_t * ki, void * d) {
	printf("[onyx-the-black-cat] Starting patching ...\n");
    printf("[onyx-the-black-cat] Finding sysent table...\n");
	// get the sysent address
	_sysent = find_sysent();
	if (_sysent == NULL) {
	    printf("[onyx-the-black-cat] Error: Cannot find sysent table\n");
		return KERN_FAILURE;
	}
	
	// save address for the real functions
	real_ptrace = (ptrace_func_t *) _sysent[SYS_ptrace].sy_call;
	real_sysctl = (sysctl_func_t *) _sysent[SYS___sysctl].sy_call;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	real_getxattr = (getxattr_func_t *) _sysent[SYS_getxattr].sy_call;
	real_fgetxattr = (fgetxattr_func_t *) _sysent[SYS_fgetxattr].sy_call;
	real_setxattr = (setxattr_func_t *) _sysent[SYS_setxattr].sy_call;
	real_fsetxattr = (fsetxattr_func_t *) _sysent[SYS_fsetxattr].sy_call;
	real_removexattr = (removexattr_func_t *) _sysent[SYS_removexattr].sy_call;
	real_fremovexattr = (fremovexattr_func_t *) _sysent[SYS_fremovexattr].sy_call;
	real_listxattr = (listxattr_func_t *) _sysent[SYS_listxattr].sy_call;
	real_flistxattr = (flistxattr_func_t *) _sysent[SYS_flistxattr].sy_call;
#endif
//	real_chmod = (chmod_func_t *)_sysent[SYS_chmod].sy_call;
//	real_open = (open_func_t *) _sysent[SYS_open].sy_call;
	
	// replace the original functions with ours, aka HIJACK
	_sysent[SYS_ptrace].sy_call = (sy_call_t *) our_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) our_sysctl;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	_sysent[SYS_getxattr].sy_call = (sy_call_t *) our_getxattr;
	_sysent[SYS_fgetxattr].sy_call = (sy_call_t *) our_fgetxattr;
	_sysent[SYS_setxattr].sy_call = (sy_call_t *) our_setxattr;
	_sysent[SYS_fsetxattr].sy_call = (sy_call_t *) our_fsetxattr;
	_sysent[SYS_removexattr].sy_call = (sy_call_t *) our_removexattr;
	_sysent[SYS_fremovexattr].sy_call = (sy_call_t *) our_fremovexattr;
	_sysent[SYS_listxattr].sy_call = (sy_call_t *) our_listxattr;
	_sysent[SYS_flistxattr].sy_call = (sy_call_t *) our_flistxattr;
#endif
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) our_chmod;
//	_sysent[SYS_open].sy_call = (sy_call_t *) our_open;

	printf("[onyx-the-black-cat] Patching is successful! Have fun...\n");
	// ALL DONE
	return KERN_SUCCESS;
}


// THE FUN ENDS HERE
kern_return_t onyx_the_black_cat_stop (kmod_info_t * ki, void * d) {
	printf("[onyx-the-black-cat] Starting unpatching ...\n");
	_sysent[SYS_ptrace].sy_call = (sy_call_t *) real_ptrace;
	_sysent[SYS___sysctl].sy_call = (sy_call_t *) real_sysctl;
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
	_sysent[SYS_getxattr].sy_call = (sy_call_t *) real_getxattr;
	_sysent[SYS_fgetxattr].sy_call = (sy_call_t *) real_fgetxattr;
	_sysent[SYS_setxattr].sy_call = (sy_call_t *) real_setxattr;
	_sysent[SYS_fsetxattr].sy_call = (sy_call_t *) real_fsetxattr;
	_sysent[SYS_removexattr].sy_call = (sy_call_t *) real_removexattr;
	_sysent[SYS_fremovexattr].sy_call = (sy_call_t *) real_fremovexattr;
	_sysent[SYS_listxattr].sy_call = (sy_call_t *) real_listxattr;
	_sysent[SYS_flistxattr].sy_call = (sy_call_t *) real_flistxattr;
#endif
//	_sysent[SYS_open].sy_call = (sy_call_t *) real_open;
//	_sysent[SYS_chmod].sy_call = (sy_call_t *) real_chmod;
	printf("[onyx-the-black-cat] Unpatching successful ...\n");
    // ALL DONE
	return KERN_SUCCESS;
}
